import optparse
from typing import Any, Iterable, Type, TypeVar, cast

from pip._internal.commands.install import InstallCommand
from pip._internal.resolution.base import BaseResolver
from pip._internal.resolution.legacy.resolver import Resolver
from pip._internal.utils.logging import setup_logging
from pip._internal.utils.temp_dir import global_tempdir_manager, tempdir_registry

__all__ = [
    "InternalLTTError",
    "PatchedInstallCommand",
    "make_pip_install_parser",
    "run",
    "new_from_similar",
    "PatchedResolverBase",
]


class InternalLTTError(RuntimeError):
    def __init__(self) -> None:
        msg = (
            "Unexpected internal ltt error. If you ever encounter this "
            "message during normal operation, please submit a bug report at "
            "https://github.com/pmeier/light-the-torch/issues"
        )
        super().__init__(msg)


class PatchedInstallCommand(InstallCommand):
    def __init__(
        self, name: str = "name", summary: str = "summary", **kwargs: Any
    ) -> None:
        super().__init__(name, summary, **kwargs)


def make_pip_install_parser() -> optparse.OptionParser:
    return cast(optparse.OptionParser, PatchedInstallCommand().parser)


def get_verbosity(options: optparse.Values, verbose: bool) -> int:
    if not verbose:
        return -1

    return cast(int, options.verbose) - cast(int, options.quiet)


def run(
    cmd: InstallCommand, args: Iterable[str], options: optparse.Values, verbose: bool
) -> int:
    with cmd.main_context():
        cmd.tempdir_registry = cmd.enter_context(tempdir_registry())
        cmd.enter_context(global_tempdir_manager())

        setup_logging(
            verbosity=get_verbosity(options, verbose),
            no_color=options.no_color,
            user_log_file=options.log,
        )

        return cast(int, cmd.run(options, list(args)))


def get_public_or_private_attr(obj: Any, name: str) -> Any:
    try:
        return getattr(obj, name)
    except AttributeError:
        try:
            return getattr(obj, f"_{name}")
        except AttributeError:
            msg = f"'{type(obj)}' has no attribute '{name}' or '_{name}'"
            raise AttributeError(msg)


T = TypeVar("T")


def new_from_similar(cls: Type[T], obj: Any, names: Iterable[str], **kwargs: Any) -> T:
    attrs = {name: get_public_or_private_attr(obj, name) for name in names}
    attrs.update(kwargs)
    return cls(**attrs)  # type: ignore[call-arg]


class PatchedResolverBase(Resolver):
    @classmethod
    def from_resolver(cls, resolver: BaseResolver) -> "PatchedResolverBase":
        return new_from_similar(
            cls,
            resolver,
            (
                "preparer",
                "finder",
                "wheel_cache",
                "upgrade_strategy",
                "force_reinstall",
                "ignore_dependencies",
                "ignore_installed",
                "ignore_requires_python",
                "use_user_site",
                "make_install_req",
                "py_version_info",
            ),
        )



import re
from typing import Any, Iterable, List, NoReturn, Optional, Text, Tuple, Union, cast

from pip._internal.index.collector import LinkCollector
from pip._internal.index.package_finder import (
    CandidateEvaluator,
    CandidatePreferences,
    LinkEvaluator,
    PackageFinder,
)
from pip._internal.models.candidate import InstallationCandidate
from pip._internal.models.link import Link
from pip._internal.models.search_scope import SearchScope
from pip._internal.req.req_install import InstallRequirement
from pip._internal.req.req_set import RequirementSet

from ..computation_backend import ComputationBackend, detect_computation_backend
from .common import (
    InternalLTTError,
    PatchedInstallCommand,
    PatchedResolverBase,
    new_from_similar,
    run,
)
from .extract import extract_dists

__all__ = ["find_links"]


def find_links(
    pip_install_args: List[str],
    computation_backend: Optional[Union[str, ComputationBackend]] = None,
    platform: Optional[str] = None,
    python_version: Optional[str] = None,
    verbose: bool = False,
) -> List[str]:
    """Find wheel links for direct or indirect PyTorch distributions with given
    properties.

    Args:
        pip_install_args: Arguments passed to ``pip install`` that will be searched for
            required PyTorch distributions
        computation_backend: Computation backend, for example ``"cpu"`` or ``"cu102"``.
            Defaults to the available hardware of the running system preferring CUDA
            over CPU.
        platform: Platform, for example ``"linux_x86_64"`` or ``"win_amd64"``. Defaults
            to the platform of the running system.
        python_version: Python version, for example ``"3"`` or ``"3.7"``. Defaults to
            the version of the running interpreter.
        verbose: If ``True``, print additional information to STDOUT.

    Returns:
        Wheel links with given properties for all required PyTorch distributions.
    """
    if isinstance(computation_backend, str):
        computation_backend = ComputationBackend.from_str(computation_backend)

    dists = extract_dists(pip_install_args)

    cmd = StopAfterPytorchLinksFoundCommand(computation_backend=computation_backend)
    pip_install_args = adjust_pip_install_args(dists, platform, python_version)
    options, args = cmd.parser.parse_args(pip_install_args)
    try:
        run(cmd, args, options, verbose)
    except PytorchLinksFound as resolution:
        return resolution.links
    else:
        raise InternalLTTError


def adjust_pip_install_args(
    pip_install_args: List[str], platform: Optional[str], python_version: Optional[str]
) -> List[str]:
    if platform is None and python_version is None:
        return pip_install_args

    if platform is not None:
        pip_install_args = maybe_add_option(
            pip_install_args, "--platform", value=platform
        )
    if python_version is not None:
        pip_install_args = maybe_add_option(
            pip_install_args, "--python-version", value=python_version
        )
    return maybe_set_required_options(pip_install_args)


def maybe_add_option(
    args: List[str],
    option: str,
    value: Optional[str] = None,
    aliases: Iterable[str] = (),
) -> List[str]:
    if any(arg in args for arg in (option, *aliases)):
        return args

    additional_args = [option]
    if value is not None:
        additional_args.append(value)
    return additional_args + args


def maybe_set_required_options(pip_install_args: List[str]) -> List[str]:
    pip_install_args = maybe_add_option(
        pip_install_args, "-t", value=".", aliases=("--target",)
    )
    pip_install_args = maybe_add_option(
        pip_install_args, "--only-binary", value=":all:"
    )
    return pip_install_args


class PytorchLinksFound(RuntimeError):
    def __init__(self, links: List[str]) -> None:
        self.links = links


class PytorchLinkEvaluator(LinkEvaluator):
    HAS_LOCAL_PATTERN = re.compile(r"[+](cpu|cu\d+)$")
    EXTRACT_LOCAL_PATTERN = re.compile(r"^/whl/(?P<local_specifier>(cpu|cu\d+))")

    @classmethod
    def from_link_evaluator(
        cls, link_evaluator: LinkEvaluator
    ) -> "PytorchLinkEvaluator":
        return new_from_similar(
            cls,
            link_evaluator,
            (
                "project_name",
                "canonical_name",
                "formats",
                "target_python",
                "allow_yanked",
                "ignore_requires_python",
            ),
        )

    def evaluate_link(self, link: Link) -> Tuple[bool, Optional[Text]]:
        output = cast(Tuple[bool, Optional[Text]], super().evaluate_link(link))
        is_candidate, result = output
        if not is_candidate:
            return output

        result = cast(Text, result)
        has_local = self.HAS_LOCAL_PATTERN.search(result) is not None
        if has_local:
            return output

        return True, f"{result}+{self.extract_computation_backend_from_link(link)}"

    def extract_computation_backend_from_link(self, link: Link) -> Optional[str]:
        match = self.EXTRACT_LOCAL_PATTERN.match(link.path)
        if match is None:
            return "any"

        return match.group("local_specifier")


class PytorchCandidatePreferences(CandidatePreferences):
    def __init__(
        self,
        *args: Any,
        computation_backend: Optional[ComputationBackend] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        if computation_backend is None:
            computation_backend = detect_computation_backend()
        self.computation_backend = computation_backend

    @classmethod
    def from_candidate_preferences(
        cls,
        candidate_preferences: CandidatePreferences,
        computation_backend: Optional[ComputationBackend] = None,
    ) -> "PytorchCandidatePreferences":
        return new_from_similar(
            cls,
            candidate_preferences,
            ("prefer_binary", "allow_all_prereleases",),
            computation_backend=computation_backend,
        )


class PytorchCandidateEvaluator(CandidateEvaluator):
    def __init__(
        self,
        *args: Any,
        computation_backend: Optional[ComputationBackend] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.computation_backend = computation_backend

    @classmethod
    def from_candidate_evaluator(
        cls,
        candidate_evaluator: CandidateEvaluator,
        computation_backend: Optional[ComputationBackend] = None,
    ) -> "PytorchCandidateEvaluator":
        return new_from_similar(
            cls,
            candidate_evaluator,
            (
                "project_name",
                "supported_tags",
                "specifier",
                "prefer_binary",
                "allow_all_prereleases",
                "hashes",
            ),
            computation_backend=computation_backend,
        )

    def get_applicable_candidates(
        self, candidates: List[InstallationCandidate]
    ) -> List[InstallationCandidate]:
        return [
            candidate
            for candidate in super().get_applicable_candidates(candidates)
            if candidate.version.local == "any"
            or candidate.version.local == self.computation_backend
        ]


class PytorchLinkCollector(LinkCollector):
    def __init__(
        self,
        *args: Any,
        url: str = "https://download.pytorch.org/whl/torch_stable.html",
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.search_scope = SearchScope.create(find_links=[url], index_urls=[])

    @classmethod
    def from_link_collector(
        cls, link_collector: LinkCollector
    ) -> "PytorchLinkCollector":
        return new_from_similar(cls, link_collector, ("session", "search_scope",),)


class PytorchPackageFinder(PackageFinder):
    _candidate_prefs: PytorchCandidatePreferences
    _link_collector: PytorchLinkCollector

    def __init__(
        self,
        *args: Any,
        computation_backend: Optional[ComputationBackend] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._candidate_prefs = PytorchCandidatePreferences.from_candidate_preferences(
            self._candidate_prefs, computation_backend=computation_backend
        )
        self._link_collector = PytorchLinkCollector.from_link_collector(
            self._link_collector
        )

    def make_candidate_evaluator(
        self, *args: Any, **kwargs: Any,
    ) -> PytorchCandidateEvaluator:
        candidate_evaluator = super().make_candidate_evaluator(*args, **kwargs)
        return PytorchCandidateEvaluator.from_candidate_evaluator(
            candidate_evaluator,
            computation_backend=self._candidate_prefs.computation_backend,
        )

    def make_link_evaluator(self, *args: Any, **kwargs: Any) -> PytorchLinkEvaluator:
        link_evaluator = super().make_link_evaluator(*args, **kwargs)
        return PytorchLinkEvaluator.from_link_evaluator(link_evaluator)

    @classmethod
    def from_package_finder(
        cls,
        package_finder: PackageFinder,
        computation_backend: Optional[ComputationBackend] = None,
    ) -> "PytorchPackageFinder":
        return new_from_similar(
            cls,
            package_finder,
            (
                "link_collector",
                "target_python",
                "allow_yanked",
                "format_control",
                "candidate_prefs",
                "ignore_requires_python",
            ),
            computation_backend=computation_backend,
        )


class StopAfterPytorchLinksFoundResolver(PatchedResolverBase):
    def _resolve_one(
        self, requirement_set: RequirementSet, req_to_install: InstallRequirement
    ) -> List[InstallRequirement]:
        self._populate_link(req_to_install)
        return []

    def resolve(
        self, root_reqs: List[InstallRequirement], check_supported_wheels: bool
    ) -> NoReturn:
        requirement_set = super().resolve(root_reqs, check_supported_wheels)
        links = [req.link.url for req in requirement_set.all_requirements]
        raise PytorchLinksFound(links)


class StopAfterPytorchLinksFoundCommand(PatchedInstallCommand):
    def __init__(
        self, computation_backend: Optional[ComputationBackend] = None, **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.computation_backend = computation_backend

    def _build_package_finder(self, *args: Any, **kwargs: Any) -> PytorchPackageFinder:
        package_finder = super()._build_package_finder(*args, **kwargs)
        return PytorchPackageFinder.from_package_finder(
            package_finder, computation_backend=self.computation_backend
        )

    def make_resolver(
        self, *args: Any, **kwargs: Any
    ) -> StopAfterPytorchLinksFoundResolver:
        resolver = super().make_resolver(*args, **kwargs)
        return StopAfterPytorchLinksFoundResolver.from_resolver(resolver)



import re
from typing import Any, List, NoReturn, cast

from pip._internal.req.req_install import InstallRequirement
from pip._internal.req.req_set import RequirementSet

from .common import InternalLTTError, PatchedInstallCommand, PatchedResolverBase, run

__all__ = ["extract_dists"]


def extract_dists(pip_install_args: List[str], verbose: bool = False) -> List[str]:
    """Extract direct or indirect required PyTorch distributions.

    Args:
        pip_install_args: Arguments passed to ``pip install`` that will be searched for
            required PyTorch distributions
        verbose: If ``True``, print additional information to STDOUT.

    Returns:
        Resolved required PyTorch distributions.
    """
    cmd = StopAfterPytorchDistsFoundInstallCommand()
    options, args = cmd.parser.parse_args(pip_install_args)
    try:
        run(cmd, args, options, verbose)
    except PytorchDistsFound as resolution:
        return resolution.dists
    else:
        raise InternalLTTError


class PytorchDistsFound(RuntimeError):
    def __init__(self, dists: List[str]) -> None:
        self.dists = dists


class StopAfterPytorchDistsFoundResolver(PatchedResolverBase):
    PYTORCH_CORE = "torch"
    PYTORCH_SUBS = ("vision", "text", "audio")

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._pytorch_dists = (
            self.PYTORCH_CORE,
            *[f"{self.PYTORCH_CORE}{sub}" for sub in self.PYTORCH_SUBS],
        )
        self._pytorch_core_pattern = re.compile(
            f"^{self.PYTORCH_CORE}(?!({'|'.join(self.PYTORCH_SUBS)}))"
        )
        self._required_pytorch_dists: List[str] = []

    def _resolve_one(
        self, requirement_set: RequirementSet, req_to_install: InstallRequirement
    ) -> List[InstallRequirement]:
        if req_to_install.name not in self._pytorch_dists:
            return cast(
                List[InstallRequirement],
                super()._resolve_one(requirement_set, req_to_install),
            )

        self._required_pytorch_dists.append(str(req_to_install.req))
        return []

    def resolve(
        self, root_reqs: List[InstallRequirement], check_supported_wheels: bool
    ) -> NoReturn:
        super().resolve(root_reqs, check_supported_wheels)
        raise PytorchDistsFound(self.required_pytorch_dists)

    @property
    def required_pytorch_dists(self) -> List[str]:
        dists = self._required_pytorch_dists
        if not dists:
            return []

        if not any(self._pytorch_core_pattern.match(dist) for dist in dists):
            dists.insert(0, self.PYTORCH_CORE)

        return dists


class StopAfterPytorchDistsFoundInstallCommand(PatchedInstallCommand):
    def make_resolver(
        self, *args: Any, **kwargs: Any
    ) -> StopAfterPytorchDistsFoundResolver:
        resolver = super().make_resolver(*args, **kwargs)
        return StopAfterPytorchDistsFoundResolver.from_resolver(resolver)



import argparse
import subprocess
import sys
from abc import ABC, abstractmethod
from os import path
from typing import Dict, List, NoReturn, Optional, Type

import light_the_torch as ltt

from .._pip.common import make_pip_install_parser
from ..computation_backend import CPUBackend

__all__ = ["make_command"]


class Command(ABC):
    @abstractmethod
    def __init__(self, args: argparse.Namespace) -> None:
        pass

    def run(self, pip_install_args: List[str]) -> None:
        self._run(pip_install_args)
        self.exit()

    @abstractmethod
    def _run(self, pip_install_args: List[str]) -> None:
        pass

    def exit(self, code: Optional[int] = None, error: bool = False) -> NoReturn:
        if code is None:
            code = 1 if error else 0
        sys.exit(code)


class GlobalCommand(Command):
    def __init__(self, args: argparse.Namespace) -> None:
        self.version = args.version

    def _run(self, pip_install_args: List[str]) -> None:
        if self.version:
            root = path.abspath(path.join(path.dirname(__file__), ".."))
            print(f"{ltt.__name__}=={ltt.__version__} from {root}")


class ExtractCommand(Command):
    def __init__(self, args: argparse.Namespace) -> None:
        self.verbose = args.verbose

    def _run(self, pip_install_args: List[str]) -> None:
        dists = ltt.extract_dists(pip_install_args, verbose=self.verbose)
        print("\n".join(dists))


class FindCommand(Command):
    def __init__(self, args: argparse.Namespace) -> None:
        self.computation_backend = args.computation_backend
        self.platform = args.platform
        self.python_version = args.python_version
        self.verbose = args.verbose

    def _run(self, pip_install_args: List[str]) -> None:
        links = ltt.find_links(
            pip_install_args,
            computation_backend=self.computation_backend,
            platform=self.platform,
            python_version=self.python_version,
            verbose=self.verbose,
        )
        print("\n".join(links))


class InstallCommand(Command):
    def __init__(self, args: argparse.Namespace) -> None:
        self.force_cpu = args.force_cpu
        self.pytorch_only = args.pytorch_only

        install_cmd = args.install_cmd
        if "{packages}" not in install_cmd:
            self.exit(error=True)
        self.install_cmd = install_cmd

        self.verbose = args.verbose

    def _run(self, pip_install_args: List[str]) -> None:
        links = ltt.find_links(
            pip_install_args,
            computation_backend=CPUBackend() if self.force_cpu else None,
            verbose=self.verbose,
        )
        cmd = self.install_cmd.format(packages=" ".join(links))
        subprocess.check_call(cmd, shell=True)

        if self.pytorch_only:
            self.exit()

        cmd = self.install_cmd.format(packages=self.collect_packages(pip_install_args))
        subprocess.check_call(cmd, shell=True)

    def collect_packages(self, pip_install_args: List[str]) -> str:
        parser = make_pip_install_parser()
        options, args = parser.parse_args(pip_install_args)
        editables = [f"-e {e}" for e in options.editables]
        requirements = [f"-r {r}" for r in options.requirements]
        return " ".join(editables + requirements + args)


COMMAD_CLASSES: Dict[Optional[str], Type[Command]] = {
    None: GlobalCommand,
    "extract": ExtractCommand,
    "find": FindCommand,
    "install": InstallCommand,
}


def make_command(args: argparse.Namespace) -> Command:
    cls = COMMAD_CLASSES[args.subcommand]
    return cls(args)



import argparse
import sys
from typing import List, Optional, Sequence, Tuple

from .._pip.common import make_pip_install_parser
from .commands import make_command

__all__ = ["main"]


def main(args: Optional[List[str]] = None) -> None:
    args = parse_args(args)
    cmd = make_command(args)

    try:
        pip_install_args = args.args
    except AttributeError:
        pip_install_args = []
    cmd.run(pip_install_args)


def parse_args(args: Optional[List[str]] = None) -> argparse.Namespace:
    if args is None:
        args = sys.argv[1:]

    parser = make_ltt_parser()
    return parser.parse_args(args)


class LTTParser(argparse.ArgumentParser):
    def parse_known_args(
        self,
        args: Optional[Sequence[str]] = None,
        namespace: Optional[argparse.Namespace] = None,
    ) -> Tuple[argparse.Namespace, List[str]]:
        args, argv = super().parse_known_args(args=args, namespace=namespace)
        if not argv:
            return args, argv

        message = (
            f"Unrecognized arguments: {', '.join(argv)}. If they were meant as "
            "optional 'pip install' arguments, they have to be passed after a '--' "
            "seperator."
        )
        self.error(message)

    @staticmethod
    def add_verbose(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--verbose",
            action="store_true",
            help=(
                "print more output to STDOUT. For fine control use -v / --verbose and "
                "-q / --quiet of the 'pip install' options"
            ),
        )

    @staticmethod
    def add_pip_install_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "args",
            nargs="*",
            help=(
                "arguments of 'pip install'. Optional arguments have to be seperated "
                "by '--'"
            ),
        )

    @staticmethod
    def add_common_arguments(parser: argparse.ArgumentParser) -> None:
        LTTParser.add_verbose(parser)
        LTTParser.add_pip_install_args(parser)


def make_ltt_parser() -> LTTParser:
    parser = LTTParser(prog="ltt")
    parser.add_argument(
        "-V",
        "--version",
        action="store_true",
        help="show light-the-torch version and path and exit",
    )

    subparsers = parser.add_subparsers(dest="subcommand", title="subcommands")
    add_ltt_install_parser(subparsers)
    add_ltt_extract_parser(subparsers)
    add_ltt_find_parser(subparsers)

    return parser


SubParsers = argparse._SubParsersAction


def add_ltt_install_parser(subparsers: SubParsers) -> None:
    parser = subparsers.add_parser(
        "install",
        description=(
            "Install PyTorch distributions from the stable releases. The computation "
            "backend  is auto-detected from the available hardware preferring CUDA "
            "over CPU."
        ),
    )
    parser.add_argument(
        "--force-cpu",
        action="store_true",
        help="disable computation backend auto-detection and use CPU instead",
    )
    parser.add_argument(
        "--pytorch-only",
        action="store_true",
        help="install only PyTorch distributions",
    )
    parser.add_argument(
        "--install-cmd",
        type=str,
        default="python -m pip install {packages}",
        help=(
            "installation command for the PyTorch distributions and additional "
            "packages. Defaults to 'python -m pip install {packages}'"
        ),
    )
    LTTParser.add_common_arguments(parser)


def add_ltt_extract_parser(subparsers: SubParsers) -> None:
    parser = subparsers.add_parser(
        "extract", description="Extract required PyTorch distributions"
    )
    LTTParser.add_common_arguments(parser)


def add_ltt_find_parser(subparsers: SubParsers) -> None:
    parser = subparsers.add_parser(
        "find", description="Find wheel links for the required PyTorch distributions"
    )
    parser.add_argument(
        "--computation-backend",
        type=str,
        help=(
            "Only use wheels compatible with COMPUTATION_BACKEND, for example 'cu102' "
            "or 'cpu'. Defaults to the computation backend of the running system, "
            "preferring CUDA over CPU."
        ),
    )
    add_pip_install_arguments(parser, "platform", "python_version")
    LTTParser.add_common_arguments(parser)


def add_pip_install_arguments(parser: argparse.ArgumentParser, *dests: str) -> None:
    pip_install_parser = make_pip_install_parser()
    option_group = pip_install_parser.option_groups[0]
    for dest in dests:
        options = [option for option in option_group.option_list if option.dest == dest]
        assert len(options) == 1
        option = options[0]

        parser.add_argument(
            *option._short_opts, *option._long_opts, help=option.help,  # type: ignore[attr-defined]
        )



import re
import subprocess
from abc import ABC, abstractmethod
from typing import Any

__all__ = [
    "ComputationBackend",
    "CPUBackend",
    "CUDABackend",
    "detect_computation_backend",
]


class ParseError(ValueError):
    def __init__(self, string: str) -> None:
        super().__init__(f"Unable to parse {string} into a computation backend")


class ComputationBackend(ABC):
    @property
    @abstractmethod
    def local_specifier(self) -> str:
        ...

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ComputationBackend):
            return self.local_specifier == other.local_specifier
        elif isinstance(other, str):
            return self.local_specifier == other
        else:
            return False

    def __hash__(self) -> int:
        return hash(self.local_specifier)

    def __repr__(self) -> str:
        return self.local_specifier

    @classmethod
    def from_str(cls, string: str) -> "ComputationBackend":
        parse_error = ParseError(string)
        string = string.lower()
        if string == "cpu":
            return CPUBackend()
        elif string.startswith("cu"):
            match = re.match(r"^cu(da)?(?P<version>[\d.]+)$", string)
            if match is None:
                raise parse_error

            version = match.group("version")
            if "." in version:
                major, minor = version.split(".")
            else:
                major = version[:-1]
                minor = version[-1]

            return CUDABackend(int(major), int(minor))
        else:
            raise parse_error


class CPUBackend(ComputationBackend):
    @property
    def local_specifier(self) -> str:
        return "cpu"


class CUDABackend(ComputationBackend):
    def __init__(self, major: int, minor: int) -> None:
        self.major = major
        self.minor = minor

    @property
    def local_specifier(self) -> str:
        return f"cu{self.major}{self.minor}"


NVCC_RELEASE_PATTERN = re.compile(r"release (?P<major>\d+)[.](?P<minor>\d+)")


def detect_computation_backend() -> ComputationBackend:
    fallback = CPUBackend()
    try:
        output = (
            subprocess.check_output(
                "nvcc --version", shell=True, stderr=subprocess.DEVNULL,
            )
            .decode("utf-8")
            .strip()
        )
        match = NVCC_RELEASE_PATTERN.findall(output)
        if not match:
            return fallback

        major, minor = match[0]
        return CUDABackend(int(major), int(minor))
    except subprocess.CalledProcessError:
        return fallback
